# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Plugin module to generate spectrogram for vibration data periodically and optionally find energy burst in there """


__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2021 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import os
import gc
import copy
import logging

from foglamp.common import logger
import filter_ingest
from foglamp.plugins.filter.specgram.specgram_handler import SpecgramHandler

_LOGGER = logger.setup(__name__, level=logging.INFO)  # WARN

_SMALL_SHOOT_DISCHARGE_ASSET = "smallShootDischarge"
_VIBRATION_TIME_SERIES_ASSET = "vibration"

KEY_DATA1 = 'Channel0'  # translate to canonical names
KEY_DATA2 = 'Channel1'
KEY_DATA3 = 'Channel2'
KEY_DATA4 = 'Channel3'
KEY_TS = 'user_ts'

FOGLAMP_DATA_DIR = os.path.join(os.getenv("FOGLAMP_ROOT"), "data")


_DEFAULT_CONFIG = {
    'plugin': {   # mandatory filter
        'description': 'filter plugin to detect need for large discharge',
        'type': 'string',
        'default': 'specgram',
        'readonly': 'true'
    },
    'enable': {    # recommended filter
        'description': 'Enable Spectrogram filter',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Enabled',
        'order': "1"
    },
    'logLevel': {
        'description': 'log level for diagnostic or error output',
        'type': 'enumeration',
        'options': ['ERROR', 'WARN', 'INFO', 'DEBUG'],
        'default': 'INFO',
        'displayName': 'Log level',
        'order': "2"
    },
    'timeseriesAsset': {
        'description': 'Name of asset that has the vibration time series',
        'type': 'string',
        'default': _VIBRATION_TIME_SERIES_ASSET,
        'displayName': 'Asset from which vibration data is read',
        'order': "3"
    },
    'smallShootDischargeAsset': {
        'description': 'Name of asset that indicates small shoot discharge length',
        'type': 'string',
        'default': _SMALL_SHOOT_DISCHARGE_ASSET,
        'displayName': 'Output asset indicating small shoot discharge length',
        'order': "4"
    },
    'useCols': {
        'description': 'Comma separated column names of timestamp (mandatory) and data channels (to process)',
        'type': 'string',
        'default': 'user_ts,Channel0,Channel1,Channel2,Channel3',
        'displayName': 'Comma separated TS and Data channel(s)',
        'order': "5"
    },
    'samplingRate': {
        'description': 'Readings per sec',
        'type': 'integer',
        'default': '8000',
        'displayName': 'Sampling rate (per sec)',
        'order': "6"
    },
    'forwardData': {
        'description': 'Forward received data down the filter chain',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Forward data',
        'order': "7"
    },
    'processingInterval': {
        'description': 'Time interval after which to generate specgram images'
                       '(time interval strings must be supported by pandas.to_timedelta API '
                       'e.g. 15s, 30m, 2hr, 1day 4hr 30m 10s)',
        'type': 'string',
        'default': '1hr',
        'displayName': 'Spectrogram generation interval',
        'order': "8"
    },
    'specgramFilesMaxAge': {
        'description': 'Number of days after which older spectrogram images may be discarded',
        'type': 'integer',
        'default': '30',
        'displayName': "Delete spectrograms images older than these many days",
        'minimum': '1',
        'order': "9"
    },
    "destDir": {
        "description": "Destination directory",
        "type": "string",
        "default": "{}/specgrams".format(FOGLAMP_DATA_DIR),
        "displayName": "Destination directory",
        "order": "10"
    },
    'filterDataStream': {
        'description': 'Whether to filter data stream here or stream is already pre-filtered as per interest',
        'type': 'enumeration',
        'options': ['Filter data stream', 'Pre-filtered data stream'],
        'default': 'Filter data stream',
        'displayName': 'Data stream filtering',
        'order': "11"
    },
    "periodicEventTimestamp": {
        "description": "UTC Timestamp (potentially in the past) when a periodic event of interest happened/would happen"
                       " For example : 2024-04-27 09:25:35.300875 ",
        "type": "string",
        "default": "2024-01-01 12:00:00.000000",
        "displayName": "Periodic event UTC timestamp",
        'validity': "filterDataStream == \"Filter data stream\"",
        "mandatory": "true",
        "order": "12"
    },
    "eventRepetitionTime": {
        "description": "The time after which the desired event repeats [eg., 16min]"
                       "(time interval strings must be supported by pandas.to_timedelta API "
                       "e.g. 15s, 30m, 2hr, 1day 4hr 30m 10s)",
        "type": "string",
        "default": "16min",
        "displayName": "Event repetition time",
        'validity': "filterDataStream == \"Filter data stream\"",
        "mandatory": "true",
        "order": "13"
    },
    "eventPreTime": {
        "description": "The amount of time to consider for collection"
                       " before the desired event [eg., 1min]"
                       "(time interval strings must be supported by pandas.to_timedelta API "
                       "e.g. 15s, 30m, 2hr, 1day 4hr 30m 10s)",
        "type": "string",
        "default": "1min",
        "displayName": "Pre event time",
        "order": "14"
    },
    "eventDuration": {
        "description": "The duration for desired event [eg., 1min]"
                       "(time interval strings must be supported by pandas.to_timedelta API "
                       "e.g. 15s, 30m, 2hr, 1day 4hr 30m 10s)",
        "type": "string",
        "default": "1min",
        "displayName": "Event duration",
        "order": "15"
    },
    "eventPostTime": {
        "description": "The amount of time to consider for collection"
                       " after the desired event [eg., 1min]"
                       "(time interval strings must be supported by pandas.to_timedelta API "
                       "e.g. 15s, 30m, 2hr, 1day 4hr 30m 10s)",
        "type": "string",
        "default": "1min",
        "displayName": "Post event time",
        "order": "16"
    },
}


def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    _LOGGER.info("plugin_info called")
    return {
        'name': 'specgram',
        'version': '3.0.0',
        'mode': 'none',
        'type': 'filter',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


LOGLEVEL = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARN': logging.WARN,
    'ERROR': logging.ERROR
}


class FilterState:
    def __init__(self, config, ingest_ref, callback):
        self.config = config
        self.ingest_ref = ingest_ref
        self.callback = callback

        self.small_shoot_asset = config['smallShootDischargeAsset']['value']
        self.ts_asset = config['timeseriesAsset']['value']
        self.forwardData = (config['forwardData']['value'] == 'true')
        _LOGGER.info("self.forwardData={}".format(self.forwardData))

        self.cols = [KEY_TS, KEY_DATA1, KEY_DATA2, KEY_DATA3, KEY_DATA4] \
            if (config['useCols']['value'] == '') else config['useCols']['value'].split(',')

        self.enabled = config['enable']['value'] == 'true'
        self.state = 0
        self.last_state = 0

        _LOGGER.setLevel(LOGLEVEL[config['logLevel']['value']])


def plugin_init(config, ingest_ref, callback):
    """ Initialise the plugin.
    Args:
        config: JSON configuration document for the plugin configuration category
        ingest_ref: reference object for ingesting readings
        callback: callback to call to ingest readings to C south service
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """

    try:
        config = copy.deepcopy(config)
        _state = FilterState(config, ingest_ref, callback)

        if _state.enabled:

            # create a new process and instantiate an rpc server

            _LOGGER.debug("config={}".format(config))
            _specgram_handler = SpecgramHandler()

            _ = _specgram_handler.initialize(config)
            config['handler'] = _specgram_handler
            config['state'] = _state
        else:
            config['handler'] = None
            config['state'] = _state

    except Exception as ex:
        _LOGGER.exception('Error in initializing specgram filter {}'.format(str(ex)))
        raise

    return config


def _clean_filter_input(data, cols):
    """ _clean_filter_input -- fix some common problems in input, stripping excess quotes, 
    Arguments:
        data : [
          dict {
              readings: {
                data_col: data,
                [ts_col: timestamp]
          }
          ...
        ]
    Returns:
        list of dicts of readings data
    """
    if type(data) == dict:
        data = [data]

    if len(data) > 0:
        ts_col = KEY_TS if len(cols) == 0 else cols[0]
        
        has_ts = ts_col in data[0]
        # if you're unlucky enough to have dates enclosed in double quotes, JSON will croak
        for d in data:
            if has_ts:
                d['readings'][ts_col] = d['readings'][ts_col].strip('"')
    
    return data


def plugin_ingest(handle, data):
    """ Returns assets to monitor for vibration data
    Args:
        handle: handle returned by the plugin initialisation call
        data: list of
        dicts of
          'readings': {'Channel1': v1, 'Channel2': v2, 'Channel3': v3, 'Channel4': v4, 'user_ts': timestamp},
          'asset': asset-name,
          'id': id,
          'ts': current-timestamp,
          'user_ts': timestamp-of-data
    Action:
        passes data to analysis algorithm, which returns:
        * width of small shoot discharge, if one is found
        * start timestamp of small shoot discharge, if found
        calls the ingest callback to pass along input data
        calls the ingest callback with a new asset if there is a valid small shoot discharge information
    Returns:
    """
    _state = handle['state']
    _LOGGER.debug("_state={}, len(data)={}".format(_state, len(data)))
        
    if _state is None or len(data) == 0:
        return  # shut down

    _LOGGER.debug("ingest: {} {} ".format(len(data), str(data[0])))
    _LOGGER.debug("convert {}".format(str(data[0]['readings'])))

    if not _state.enabled:  # forward data if filter is disabled
        _LOGGER.debug("Forwarding original {} readings as filter is disabled".format(len(data)))
        filter_ingest.filter_ingest_callback(_state.callback, _state.ingest_ref, data)
        del data
        return

    # pass data up to upper levels
    if _state.forwardData:
        _LOGGER.debug("Forwarding original {} readings".format(len(data)))
        filter_ingest.filter_ingest_callback(_state.callback, _state.ingest_ref, data)

    # if incoming reading list has a mix of assets, process only the matching readings & forward non-matching readings
    non_matching_asset_readings = [d for d in data if d['asset'] != _state.ts_asset]
    _LOGGER.debug("Forwarding {} readings with non-matching asset".format(len(non_matching_asset_readings)))
    if not _state.forwardData and len(non_matching_asset_readings):
        filter_ingest.filter_ingest_callback(_state.callback, _state.ingest_ref, non_matching_asset_readings)
    del non_matching_asset_readings

    _data = [d for d in data if d['asset'] == _state.ts_asset]
    data = _data
    _LOGGER.debug("Processing {} readings with matching asset".format(len(data)))

    _specgram_handler = handle['handler']

    # run backend and generate spectrogram
    width, small_shoot_start_time = _specgram_handler.generate_specgram(data)
    _LOGGER.debug("plugin_ingest (RPC call) returned: width={}, small_shoot_start_time={}"
                  .format(width, small_shoot_start_time))

    if width > 0:  # valid
        # _LOGGER.debug(lsos())
        # _state.state = _result

        # must have 1-1 assets we add vs assets we receive
        if 'timestamp' in data[-1]:
            timestamp = data[-1]['timestamp']
        else:
            timestamp = data[-1]['ts']

        asset_dict = {'asset': _state.small_shoot_asset,
                      'readings': {'small_shoot_discharge_width': width,
                                   'small_shoot_discharge_start_time': small_shoot_start_time.strftime('%Y-%m-%d '
                                                                                                       '%H:%M:%S.%f')},
                      'timestamp': timestamp
                      }

        # pass new reading(s) data up to upper levels
        _LOGGER.info("sending reading: {}".format([asset_dict]))
        filter_ingest.filter_ingest_callback(_state.callback, _state.ingest_ref, [asset_dict])
        del asset_dict


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin
    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """

    new_handle = plugin_init(new_config, handle['state'].ingest_ref, handle['state'].callback)
    return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        plugin shutdown
    """
    _state = handle['state']

    _LOGGER.info('specgram plugin shut down.')

    _specgram_handler = handle['handler']

    if _state.enabled:
        _specgram_handler.shutdown()

    _specgram_handler = None
    _state = None

    handle['handler'] = _specgram_handler
    handle['state'] = _state

    _LOGGER.info("specgram::plugin_shutdown() completed")

    # remove _LOGGER and its connection to syslogd
    del _LOGGER.manager.loggerDict[__name__]
    del _LOGGER
    _LOGGER = None
