# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Server functions for plugin module to detect need for large discharge (LD) """

__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2021 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import os
import gc
import time
import copy
import datetime
import logging

import numpy as np
import pandas as pd

import matplotlib
matplotlib.use('Agg')
import multiprocessing
from concurrent.futures import ProcessPoolExecutor
import matplotlib.pyplot as plt
from matplotlib import figure
# from matplotlib import colors
from matplotlib.ticker import FuncFormatter  # (AutoMinorLocator, MultipleLocator, FuncFormatter)

from scipy import signal

from foglamp.common import logger, utils


# canonical data field names
KEY_DATA1 = 'Channel0'
KEY_DATA2 = 'Channel1'
KEY_DATA3 = 'Channel2'
KEY_DATA4 = 'Channel3'
KEY_TS = 'user_ts'

# very special: if we are passed in a default logger destination use that
if hasattr(logger, 'FOGLAMP_DESTINATION_ENV') and (logger.FOGLAMP_DESTINATION_ENV in os.environ):
    logger.set_default_destination(int(os.environ[logger.FOGLAMP_DESTINATION_ENV]))
else:
    utils.eprint("NO DEST ENV")

_LOGGER = logger.setup(__name__, level=logging.INFO)

LOGLEVEL = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARN': logging.WARN,
    'ERROR': logging.ERROR
}


def plot_specgram(logger, time_series, title, sampling_rate,
                  save_dir, send_end, start_ts=None, end_ts=None):
    """ Plots spectrogram image for given time series
    Args:
        logger: Logger object to write logs.
        time_series: Vibration data for a single channel
        title: Title to use in spectrogram images
        sampling_rate : The sampling rate used to sample data.
        save_dir: The directory where the images are to be saved.
        send_end: A process pip to send the calculated results.
        start_ts: Timestamp at beginning of time series
        end_ts: Timestamp at end of time series
    Returns:
        start_x: Offset of beginning of small shoot (energy burst) from start of time series (in secs)
        stop_x: Offset of end of small shoot (energy burst) from start of time series (in secs)
    """
    logger.debug("Entered plot_specgram(): signal duration={} secs".format((end_ts - start_ts).total_seconds()))
    # nfft = 8 * 1024
    fs = f_s = sampling_rate

    if (end_ts - start_ts).total_seconds() < 10.0:
        _nperseg = 256
        _nfft = 1024
        fig_width = 20
        fig_height = 5
    else:
        _nperseg = 256
        _nfft = 256
        fig_width = 26
        fig_height = 5
    
    try:
        freqs, times, Sx = signal.spectrogram(time_series, fs=f_s, window='hann',
                                          nperseg=_nperseg,
                                          # nperseg=1024//16,  # defaults to 256 if window is str
                                          # noverlap=(1024//16)//8,  # If None, noverlap = nperseg // 8
                                          nfft=_nfft,
                                          detrend='linear',
                                          scaling='spectrum')  # computing the power spectrum

        mn, mx = np.min(Sx), np.max(Sx)
        logger.info("Sx.shape={}, Sx.min/max = {}..{}".format(Sx.shape, mn, mx))
        logger.info("Sx.log-min/max = {}..{}".format(np.log10(mn), np.log10(mx)))
        #     Sx = np.clip(Sx, 5e-8, 0.16)
        #     Sx = np.array(Sx)
        #     Sx[Sx < 5e-10] = 1e-20
        logger.info("fs={}, nfft={}, FFT bin size = {} Hz".format(fs, _nfft, freqs[1]))
        
        f = plt.figure(figsize=(fig_width, fig_height))
        fig,ax = plt.subplots(1,figsize=(fig_width, fig_height))
        plt.pcolormesh(times, freqs, 10 * np.log10(Sx), vmin=-100, vmax=0, shading='gouraud', 
                       cmap='rainbow')  # plasma, viridis, rainbow, hsv, coolwarm, Spectral
        plt.colorbar()

        ax.axis('tight')
        ax.set_ylabel('Frequency [Hz]')
        ax.set_xlabel('Time [s]')

        if start_ts:
            def map_x(x, pos):
                """ The two args are the value and tick position """
                return (start_ts + pd.Timedelta("{} sec".format(x))).strftime("%H:%M:%S.%f")

            formatter = FuncFormatter(map_x)
            plt.rcParams.update({'xtick.labelsize': 'large', 'ytick.labelsize': 'large'})
            ax.xaxis.set_major_formatter(formatter)

        y_limit = int(f_s / 2)
        ax.set_ylim(0.0, y_limit)

        ax.set_title(title)
        # self.logger.debug("ax.get_xlim()={}".format(ax.get_xlim()))

        # self.logger.debug("Before log, Sx.shape={}".format(Sx.shape))
        Sx = np.sum(Sx, axis=0)
        logger.debug("After log and column wise sum, Sx.shape={}, mean={}, min={}"
                        .format(Sx.shape, np.mean(Sx), np.min(Sx)))

        mn = np.mean(Sx)
        Sx[Sx < mn * 3 / 4] = 0  # mn*2

        from scipy.signal import find_peaks, peak_widths
        peaks, _ = find_peaks(Sx, height=mn * 4)

        width, width_height, left_ips, right_ips = peak_widths(Sx, peaks, rel_height=1)
        logger.debug("width={}, width_heights={}, left_ips={}, right_ips={}"
                        .format(width, width_height, left_ips, right_ips))
        if len(width) == 0:
            # plt.show()
            today = str(datetime.date.today())
            os.makedirs(os.path.join(save_dir, today), exist_ok=True)
            img_filepath = os.path.join(save_dir, today,
                                     title.replace(" : ", "_").replace(":", "_") + '.png')
            plt.savefig(img_filepath,
                        bbox_inches='tight', facecolor='w')
            logger.debug("1. Specgram saved to {}".format(img_filepath))

            plt.cla()  # Clear the current axes
            plt.clf()  # Clear the current figure
            plt.close('all')  # Closes all the figure windows
            plt.close(f)

            logger.info("Unable to find any peaks, skipping...")
            send_end.send((0, 0))
            return

        idx = np.argmax(width)
        width = width[idx]
        width_height = width_height[idx]
        left_ips = left_ips[idx]
        right_ips = right_ips[idx]
        logger.debug("max width at index {}: width={}, height={}, range={}..{}"
                        .format(idx, width, width_height, left_ips, right_ips))

        start_x = left_ips / Sx.shape[0] * (end_ts - start_ts).total_seconds()
        stop_x = right_ips / Sx.shape[0] * (end_ts - start_ts).total_seconds()
        logger.debug("start_x={} s, stop_x={} s".format(start_x, stop_x))

        plt.axvline(x=start_x, linestyle='--', color='black')
        plt.axvline(x=stop_x, linestyle='--', color='black')

        # if (end_ts - start_ts).total_seconds() < 10.0:
        today = str(datetime.date.today())
        os.makedirs(os.path.join(save_dir, today), exist_ok=True)
        img_filepath = os.path.join(save_dir, today, title.replace(" : ", "_").replace(":", "_") + '_found.png')
        plt.savefig(img_filepath,
                    bbox_inches='tight', facecolor='w')
        logger.debug("2. Specgram saved to {}".format(img_filepath))
        # plt.show()
        plt.cla()  # Clear the current axes
        plt.clf()  # Clear the current figure
        plt.close('all')  # Closes all the figure windows
        plt.close(f)

        idx = np.argmax(Sx)
        t = (end_ts - start_ts) / Sx.shape[0] * idx
        logger.debug("idx={}, t={:.0f} ms".format(idx, t.total_seconds() * 1000))
        logger.debug("Max value is {} at index {}={}".format(Sx[idx], idx, (start_ts + t).strftime("%H:%M:%S.%f")))
        logger.info("small shoot is from offset {} .. {} sec\n".format(start_x, stop_x))

        send_end.send((start_x, stop_x))
        return
    except Exception as ex:
        logger.exception('Error in plot_specgram(): {}'.format(str(ex)))
        raise


def evaluate(channel, start_second, end_second, segment_len_secs, df2, logger, sampling_rate,
             destDir, main_send_end):
    """
    Function for evaluation of peaks in vibration data and finally spot the main energy burst.
    :param channel: Channel Name
    :type channel: string
    :param start_second: start time of the signal.
    :type start_second: timestamp object
    :param end_second: end time of the signal.
    :type end_second: timestamp object
    :param segment_len_secs: the segment lengths in which signal is broken and then peaks are evaulated.
    :type segment_len_secs: integer
    :param df2: The main vibration signal.
    :type df2: numpy array
    :param logger: The logger object.
    :type logger: object
    :param sampling_rate: The sampling rate of signal.
    :type sampling_rate: integer
    :param destDir: The directory where plots are to be saved.
    :type destDir: string
    :param main_send_end: A process pipe where the result of evaluation is to be sent.
    :type main_send_end: A process pipe
    :return: a tuple of (width of energy burst , start of energy burst)
    :rtype: a tuple
    """
    logger.debug("inside evaluate ")

    fs = sampling_rate

    for start_sec in np.arange(0, segment_len_secs, segment_len_secs):
        title = '{}-{}_{}_spectrogram' \
            .format(df2.index[int(fs * start_sec)].strftime("%H:%M:%S.%f"),
                    df2.index[int(fs * (start_sec + segment_len_secs)) - 1].strftime("%H:%M:%S.%f"),
                    channel)

        logger.debug("title={}".format(title))
        data = df2[channel][int(fs * start_sec):int(fs * (start_sec + segment_len_secs))]

        recv_end, send_end = multiprocessing.Pipe(False)
        p = multiprocessing.Process(target=plot_specgram, name="acc_plot", args=(logger, data, title,
                                                                sampling_rate, destDir,
                                                                send_end,
                                                                df2.index[int(fs * start_sec)],
                                                                df2.index[int(fs *
                                                                              (start_sec + segment_len_secs))
                                                                          - 1]))
        p.start()
        p.join()
        start_t, stop_t = recv_end.recv()

        if start_t == 0 and stop_t == 0:
            main_send_end.send((start_t, stop_t))
            return

        width = stop_t - start_t
        start_ts = df2.index[int(fs * (start_second + start_t))]
        stop_ts = df2.index[int(fs * (start_second + stop_t))]
        logger.debug("{}: start_t={:.3f} stop_t={:.3f}, width={:.3f}".format(channel, start_t, stop_t, width))
        logger.debug("{}: start_ts={}, stop_ts={}".format(channel, start_ts, stop_ts))

        logger.debug("{}: Iter 1: small shoot is from {} .. {}, width={:.0f} msec"
                          .format(channel, df2.index[int(fs * (start_second + start_t))],
                                  df2.index[int(fs * (start_second + stop_t))], width * 1000))
        start_sec = max(start_second, start_second + start_t - 2)  # 2*width)
        end_sec = min(start_sec + segment_len_secs, start_second + stop_t + 2, end_second)  # 2*width)
        logger.debug(
            "start_sec={}, end_sec={}, segment_len_secs={} -> {}, int(fs*(start_sec+segment_len_secs)-1)={}"
                .format(start_sec, end_sec, segment_len_secs, end_sec - start_sec,
                        int(fs * (start_sec + segment_len_secs) - 1)))
        segment_len_secs = end_sec - start_sec
        data = df2[channel][int(fs * start_sec):int(fs * (start_sec + segment_len_secs))]
        logger.debug("\n\n")

        title = '{}-{}_{}_spectrogram' \
            .format(df2.index[int(fs * start_sec)].strftime("%H:%M:%S.%f"),
                    df2.index[int(fs * (start_sec + segment_len_secs)) - 1].strftime("%H:%M:%S.%f"),
                    channel)

        recv_end, send_end = multiprocessing.Pipe(False)
        p = multiprocessing.Process(target=plot_specgram, args=(logger, data, title,
                                                                sampling_rate, destDir,
                                                                send_end,
                                                                df2.index[int(fs * start_sec)],
                                                                df2.index[int(fs *
                                                                              (start_sec + segment_len_secs))
                                                                          - 1]))
        p.start()
        p.join()
        start_t, stop_t = recv_end.recv()

        if start_t == 0 and stop_t == 0:
            main_send_end.send((start_t, stop_t))
            return

        width = stop_t - start_t
        logger.debug("{}: Iter 2: small shoot is from {} .. {}, width={:.0f} msec"
                          .format(channel, df2.index[int(fs * (start_sec + start_t))],
                                  df2.index[int(fs * (start_sec + stop_t))], width * 1000))
        logger.debug("\n====================================\n")

        main_send_end.send((width, df2.index[int(fs * (start_sec + start_t))]))
        return


class SpecgramHandler:

    def __init__(self):
        """ Initialize class internals """

        MAX_FRAMES = 10  # max number of frames that can be buffered

        self.logger = _LOGGER

        self.handle = dict()

        # data are input into an in-memory frame
        self.df = pd.DataFrame()
        self.num_rows_processed = 0
        self.num_input_rows = 0
        self.frame = None
        self.max_frames = MAX_FRAMES

        self.forwardData = False
        self.valid = False
        self.filterDisabledMsgTs = None

        self.sampling_rate = 8000
        self.cols = [KEY_TS, KEY_DATA1, KEY_DATA2, KEY_DATA3, KEY_DATA4]

        self.LocalTimeZone = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
        self.logger.info("self.LocalTimeZone={}".format(self.LocalTimeZone))

        self.destDir = os.path.join('/tmp', 'specgram_plots')
        self.specgramFilesMaxAge = None

        self.periodicEventTimestamp = None
        self.eventRepetitionTime = None
        self.eventPreTime = None
        self.eventDuration = None
        self.eventPostTime = None
        self.processingInterval = None
        self.dataCollectionDuration = None
        self.lastEventSearchBaseTime = None
        self.nextEventTimePre = None

        # self.lastProcessedDataTS = pd.Timestamp('2000-01-01 00:00:00.000000+00:00')

    def getNextDataCollectionTime(self, lastEventSearchBaseTime):
        """ Calculates time when data collection is to be done next
                Args:
                    lastEventSearchBaseTime: Time that was used as base when searching for data collection time, last
                    time around
                Returns:
                    nextEventTimePre: Time when data collection is to be done next
        """
        if not self.filterDataStream:
            return "N.A"
        
        self.logger.debug("lastEventSearchBaseTime={}, self.periodicEventTimestamp={}".format(lastEventSearchBaseTime, self.periodicEventTimestamp))
        time_diff = (lastEventSearchBaseTime - self.periodicEventTimestamp).total_seconds()  # time since given event TS
        self.logger.debug("time_diff={} sec".format(time_diff))
        self.logger.debug("eventRepetitionTime={} sec".format(self.eventRepetitionTime))
        timeSinceLastEvent = time_diff % self.eventRepetitionTime

        nextEventTimeSec = self.eventRepetitionTime - timeSinceLastEvent
        nextEventTimeSecPre = nextEventTimeSec - self.eventPreTime
        if nextEventTimeSecPre < 0:
            nextEventTimeSecPre = nextEventTimeSecPre + self.eventRepetitionTime
        self.logger.debug("timeSinceLastEvent={} sec, nextEventTimeSec={} sec, nextEventTimeSecPre={} sec"
                          .format(timeSinceLastEvent, nextEventTimeSec, nextEventTimeSecPre))

        nextEventTimePre = lastEventSearchBaseTime + pd.to_timedelta("{} sec".format(nextEventTimeSecPre))
        return nextEventTimePre

    def deleteAgedFiles(self):
        """ Delete spectrogram image folder(s) which are too old as per filter config
                Args:
                Returns:
                    None
        """
        now = time.time()

        for f in os.listdir(self.destDir):
            path = os.path.join(self.destDir, f)
            if os.path.isdir(path) and os.stat(path).st_mtime < (now - self.specgramFilesMaxAge):
                self.logger.info("Directory {} is older than given date, deleting it...".format(f))
                os.system("rm -rf " + path)

    def initialize(self, handle):
        """ Initializes class state using passed config
                Args:
                    handle: config dict for the filter
                Returns:
                    handle: config dict for the filter
        """
        self.logger.info('Initialize Specgram Handler')
        self.logger.setLevel(LOGLEVEL[handle['logLevel']['value']])

        handle = copy.deepcopy(handle)

        # initialize data and common settings
        _sampling_rate = handle['samplingRate']['value']
        if _sampling_rate.endswith('khz'):
            self.sampling_rate = int(_sampling_rate[:-3]) * 1000
        else:
            self.sampling_rate = int(_sampling_rate)

        self.small_shoot_asset = handle['smallShootDischargeAsset']['value']

        self.forwardData = (handle['forwardData']['value'] == 'true')
        # self.inference_gap = pd.to_timedelta(handle['evaluationInterval']['value'])

        self.filterDataStream = (handle['filterDataStream']['value'] == 'Filter data stream')
        self.logger.info("SpecgramHandler::initialize(): handle['filterDataStream']['value']={}, self.filterDataStream={}"
                            .format(handle['filterDataStream']['value'], self.filterDataStream))

        self.cols = handle['useCols']['value'].split(',')
        
        self.destDir = handle['destDir']['value']
        self.logger.debug("self.specgramFilesMaxAge = {}days".format(handle['specgramFilesMaxAge']['value']))
        self.specgramFilesMaxAge = int(handle['specgramFilesMaxAge']['value'])*24*60*60  # convert from days to seconds
        self.logger.debug("self.specgramFilesMaxAge={} secs".format(self.specgramFilesMaxAge))

        if self.filterDataStream:
            self.periodicEventTimestamp = pd.to_datetime(handle['periodicEventTimestamp']['value'], errors='coerce')
            self.eventRepetitionTime = pd.to_timedelta(handle['eventRepetitionTime']['value'], errors='coerce')
        else:
            self.periodicEventTimestamp = pd.NaT
            self.eventRepetitionTime = pd.NaT
        
        self.eventPreTime = pd.to_timedelta(handle['eventPreTime']['value'], errors='coerce')
        self.eventDuration = pd.to_timedelta(handle['eventDuration']['value'], errors='coerce')
        self.eventPostTime = pd.to_timedelta(handle['eventPostTime']['value'], errors='coerce')
        self.processingInterval = pd.to_timedelta(handle['processingInterval']['value'], errors='coerce')
        
        if self.filterDataStream and pd.isnull(self.periodicEventTimestamp):
            self.logger.error("BackendServer: plugin_init(): Unable to process periodicEventTimestamp ['{}']".format(
                handle['periodicEventTimestamp']['value']))
            self.logger.error("Specgram filter disabled. Fix the filter config")
            self.valid = False
            return None
        else:
            self.periodicEventTimestamp = self.periodicEventTimestamp.tz_localize(tz='UTC')

        if self.filterDataStream and pd.isnull(self.eventRepetitionTime):
            self.logger.error("BackendServer: plugin_init(): Unable to process eventRepetitionTime ['{}']".format(
                handle['eventRepetitionTime']['value']))
            self.logger.error("Specgram filter disabled. Fix the filter config")
            self.valid = False
            return None
        else:
            self.eventRepetitionTime = self.eventRepetitionTime.total_seconds()

        if pd.isnull(self.eventPreTime):
            self.logger.error("BackendServer: plugin_init(): Unable to process eventPreTime ['{}']".format(
                handle['eventPreTime']['value']))
            self.logger.error("Specgram filter disabled. Fix the filter config")
            self.valid = False
            return None
        else:
            self.eventPreTime = self.eventPreTime.total_seconds()

        if pd.isnull(self.eventDuration):
            self.logger.error("BackendServer: plugin_init(): Unable to process eventDuration ['{}']".format(
                handle['eventDuration']['value']))
            self.logger.error("Specgram filter disabled. Fix the filter config")
            self.valid = False
            return None
        else:
            self.eventDuration = self.eventDuration.total_seconds()

        if pd.isnull(self.eventPostTime):
            self.logger.error("BackendServer: plugin_init(): Unable to process eventPostTime ['{}']".format(
                handle['eventPostTime']['value']))
            self.logger.error("Specgram filter disabled. Fix the filter config")
            self.valid = False
            return None
        else:
            self.eventPostTime = self.eventPostTime.total_seconds()

        if pd.isnull(self.processingInterval):
            self.logger.error("BackendServer: plugin_init(): Unable to process processingInterval ['{}']".format(
                handle['processingInterval']['value']))
            self.logger.error("Specgram filter disabled. Fix the filter config")
            self.valid = False
            return None
        else:
            self.processingInterval = self.processingInterval.total_seconds()

        self.dataCollectionDuration = self.eventPreTime + self.eventDuration + self.eventPostTime
        # calculate next collection start time
        self.lastEventSearchBaseTime = pd.Timestamp.now().tz_localize(tz=self.LocalTimeZone).tz_convert('UTC') if self.filterDataStream else "N.A."
        self.nextEventTimePre = self.getNextDataCollectionTime(self.lastEventSearchBaseTime) if self.filterDataStream else "N.A."
        self.logger.debug("Searched with lastEventSearchBaseTime={}, next event pre occurs at {} and data collection "
                          "period is {} secs "
                          .format(self.lastEventSearchBaseTime, self.nextEventTimePre, self.dataCollectionDuration))

        # initialize ML-specific values
        # handle = self.ml_plugin_init(handle)

        self.collectionStart = None
        self.collectionStop = None

        self.handle = handle
        self.valid = True
        return handle


    def shutdown(self):
        """ Shutdown routine for this backend server
                Args:
                Returns:
                    None
        """
        self.logger.warning("BackendServer: SHUTTING DOWN")
        
        # remove _LOGGER and its connection to syslogd
        del self.logger.manager.loggerDict[__name__]
        del self.logger
        self.logger = None

        self.valid = False

    def generate_specgram(self, data):
        """ Processes incoming vibration data; Either accumulates or ignores data as per configured timespec. Also
        creates spectrogram images if accumulation ends
        Args:
            data: Reading data
        Returns:
            small shoot width: width in seconds, if a small shoot is found, else 0
            small shoot start time: start time of small shoot if found, else 0
        """
        self.logger.debug("BackendServer: generate_specgram")
        # handle = self.handle

        if not self.valid:
            # Output "filter is disabled" message every 2 mins
            if not self.filterDisabledMsgTs:
                self.filterDisabledMsgTs = pd.Timestamp.now() - pd.to_timedelta("2 min")
            if pd.Timestamp.now() - self.filterDisabledMsgTs >= pd.to_timedelta("2 min"):
                _LOGGER.info("Not processing data since filter is disabled")
                self.filterDisabledMsgTs = pd.Timestamp.now()
            
            del data
            return 0, 0

        try:
            # self.logger.debug("srvev: " + str(data))

            if not self.accumulate_data(data):
                return 0, 0
            else:
                # Collection has stopped; update self.nextEventTimePre
                prevLastEventSearchBaseTime = self.lastEventSearchBaseTime
                self.lastEventSearchBaseTime = self.lastEventSearchBaseTime + \
                                               pd.to_timedelta("{} sec".format(
                                                              self.processingInterval))
                self.logger.info("lastEventSearchBaseTime updated from {} to {}"
                                 .format(prevLastEventSearchBaseTime, self.lastEventSearchBaseTime))
                self.nextEventTimePre = self.getNextDataCollectionTime(
                    self.lastEventSearchBaseTime)
                self.logger.info("Next event pre occurs at {} and data collection period is {} secs"
                                 .format(self.nextEventTimePre, self.dataCollectionDuration))

            self.logger.debug("Generating spectrograms")
            ret_val = self.process_vibration_data() # return timestamps also where peak is found
            self.logger.debug("Got return value {} ".format(ret_val))

            # delete older directories, if present
            self.deleteAgedFiles()
            self.logger.info("Deleted everything in old directories.")
            del data
            del self.df
            self.logger.debug("the list of objects before collection {}".format(len(gc.get_objects())))
            gc.collect()
            self.df = pd.DataFrame()
            data = []
            self.logger.debug("the list of objects after collection {}".format(len(gc.get_objects())))

            # report max width result
            width = - 1
            small_shoot_start_time = 0
            for val in ret_val:
                if val[0] > width:
                    width = val[0]
                    small_shoot_start_time = val[1]

            # self.logger.info("Process prepared frame for generating spectrogram and finding energy burst in there.")
            return width, small_shoot_start_time

        except Exception as ex:
            self.logger.exception('Error in BackendServer plugin_ingest {}'.format(str(ex)))
            self.valid = False
            raise

    def process_vibration_data(self):
        """ Plots spectrogram image for accumulated time series
        Args:
        Returns:
            Array of pairs (one for each channel), which contains:
                width: Width of small shoot discharge
                start_time: Start time of small shoot discharge
        """
        df2 = self.df
        channels = list(df2.columns)
        self.logger.debug("----> process_vibration_data(): df2.index[0]={}, df2.index[-1]={}, channels={}"
                                 .format(df2.index[0], df2.index[-1], channels))
        # self.logger.debug("df2.head()=\n{}".format(df2.head()))

        fs = self.sampling_rate
        start_second = 0
        end_second = df2.shape[0] / fs
        segment_len_secs = end_second - start_second
        self.logger.debug("start_second={}, end_second={}".format(start_second, end_second))

        ret_val = []
        combined_data = []

        for channel in channels:
            recv_end, send_end = multiprocessing.Pipe(False)

            self.logger.debug("Calling evaluate for channel {}".format(channel))

            p = multiprocessing.Process(target=evaluate, name="acc_evaluate", args=(channel, start_second, end_second,
                                                                                    segment_len_secs, df2,
                                                                                    self.logger,
                                                                                    self.sampling_rate,
                                                                                    self.destDir, send_end))
            self.logger.debug("starting process")
            p.start()
            self.logger.debug("joining process")
            p.join()
            # combined_data.append()
            w, t = recv_end.recv()
            ret_val.append((w, t))

        return ret_val

    def accumulate_data(self, data):
        """ Accumulates incoming vibration data as per configured timespec
        Args:
            data: Reading data
        Returns:
            bool: indicating whether accumulation has completed
        """

        if len(data) == 0:
            return False
        
        self.logger.debug("accumulate_data(): len(data)={}, data[0]={}".format(len(data), data[0]))

        # t1 = datetime.datetime.now()

        _input = pd.DataFrame(data[0])
        firstTS = pd.to_datetime(_input['user_ts'][0])  # .tz_localize(tz='UTC')  # TS in first reading
        _input = pd.DataFrame(data[-1])
        lastTS = pd.to_datetime(_input['user_ts'][0])  # .tz_localize(tz='UTC')  # TS in first reading

        if not self.filterDataStream and self.df.shape[0]==0:  # pre-filtered stream and first chunk of data
            self.collectionStart = firstTS
            self.lastEventSearchBaseTime = firstTS
        
        if self.filterDataStream:
            self.collectionStart = self.nextEventTimePre

        self.collectionStop = self.collectionStart + pd.to_timedelta("{} sec".format(self.dataCollectionDuration))

        self.logger.debug("firstTS={}, lastTS={}, self.collectionStart={}, self.collectionStop={}"
                          .format(firstTS, lastTS, self.collectionStart, self.collectionStop))

        # Check intersection
        if firstTS > self.collectionStop or self.collectionStart > lastTS:
            self.logger.debug("No intersection")

            # it is possible that there is no intersection, but data stream is already past the collectionStop time;
            # in that case, process accumulated data
            if firstTS > self.collectionStop:
                self.logger.debug("Stream is already past the collectionStop time, processing it...")
            else:
                return False

        if self.collectionStop > firstTS:

            # find intersection
            fromTs = max(firstTS, self.collectionStart)
            toTs = min(lastTS, self.collectionStop)
            self.logger.debug("Intersection from {} to {}".format(fromTs, toTs))

            t1 = datetime.datetime.now()
            required_dict_list = [s['readings'] for s in data]
            required_ts_list = [s['user_ts'] for s in data]
            _input_d = pd.DataFrame(required_ts_list, columns=['user_ts_orig'])
            _input_r = pd.DataFrame.from_records(required_dict_list)
            _input = pd.concat([_input_r, _input_d], axis=1)
            input_cols = list(_input.columns)
            input_cols.remove('user_ts_orig')
            extra_cols = set(input_cols) - set(self.cols)
            self.logger.debug(
                "input_cols={}, self.cols={}, extra_cols = {}".format(input_cols, self.cols, extra_cols))
            if len(extra_cols):
                _input.drop(labels=list(extra_cols), axis=1, inplace=True)

            _input.index = pd.to_datetime(_input['user_ts_orig'])
            _input = _input[(_input.index >= fromTs) & (_input.index <= toTs)]  # TODO: Maybe move this filtering above
            self.logger.info("Accumulating data from {} to {}".format(_input.index[0], _input.index[-1]))

            lastReceivedTS = _input.index[-1]
            labels_to_remove = ['user_ts_orig']
            if 'user_ts' in _input.columns:
                labels_to_remove.append('user_ts')
            _input.drop(labels=labels_to_remove, axis=1, inplace=True)

            t2 = datetime.datetime.now()
            delta = t2 - t1
            delta = int(delta.total_seconds() * 1000 * 1000)  # in usec
            self.logger.debug("df processing took {:.0f} msec for {} readings, time/reading={:.0f} us"
                                     .format(delta / 1000, _input.shape[0], delta / _input.shape[0]))

            # self.logger.debug("lastProcessedDataTS={}, inference_gap={}, currTS={}, so processing this data frame"
            #                   .format(self.lastProcessedDataTS, self.inference_gap, currTS))

            self.logger.debug("*** Appending data with TS={}, after data with TS={}"
                              .format(_input.index[0], self.df.index[-1] if self.df.shape[0] else "N.A."))
            _len = _input.shape[0]
            self.df = pd.concat([self.df, _input])
            self.logger.debug("Inside acc The memory self.df {}  memory is {}".format(self.df.shape,
                                                                                     self.df.memory_usage()))

            self.logger.debug("### Last TS stored={}, collection stops at {}".format(lastReceivedTS,
                                                                                     self.collectionStop))

            if self.filterDataStream:
                if toTs == lastTS:  # last data row in input stream was selected, so need to wait for more data
                    self.logger.debug("### Got {} rows to process, not processing yet".format(self.df.shape[0]))
                    self.logger.debug("### Last TS stored={}, collection stops at {}"
                                    .format(self.df.index[-1], self.collectionStop))
                    return False  # don't run model inference if collection has not ended
            else:
                self.logger.debug("lastTS={}, self.collectionStop={}".format(lastTS, self.collectionStop))

                time_diff_abs = (abs(self.collectionStop - lastTS)).total_seconds()
                # sometimes Notification service doesn't deliver all data in conditional forwarding setting, 
                # it might deliver a few 10-100s of ms worth of data less; allow upto 1% less data
                time_diff_eps = self.dataCollectionDuration/100.0
                
                self.logger.debug("******* time gap between last data timestamp and collection stop time: time_diff_abs={}, time_diff_eps={}"
                                            .format(time_diff_abs, time_diff_eps))
                
                if time_diff_abs < time_diff_eps:
                    self.logger.info("### Pre-filtered stream: data collection finished: from {} to {}".format(self.df.index[0], self.df.index[-1]))
                    return True
                else:
                    self.logger.debug("### Pre-filtered stream: data collection is not finished yet")
                    return False

            self.logger.info(
                "### Data collection finished: from {} to {}".format(self.df.index[0], self.df.index[-1]))

        else:
            self.logger.info("No new data to add, just processing already collected data")

        self.logger.debug(
            "Inside finish of accumulate data The memory self.df {}  memory is {}".format(self.df.shape,
                                                                                          self.df.memory_usage()))
        return True
